@@ -11,12 +11,11 @@ import logging
 import time
 import weakref
 from collections import defaultdict
-from functools import wraps, partial
+from functools import partial, wraps
 from typing import overload
 
 import flask
 import flask.wrappers
-
 from limits.errors import ConfigurationError
 from limits.storage import MemoryStorage, Storage, storage_from_string
 from limits.strategies import STRATEGIES, RateLimiter
@@ -26,17 +25,18 @@ from .constants import MAX_BACKEND_CHECKS, ConfigVars, ExemptionScope, HeaderNam
 from .errors import RateLimitExceeded
 from .manager import LimitManager
 from .typing import (
+    Callable,
+    Dict,
+    List,
+    Optional,
     P,
     R,
-    Optional,
-    List,
-    Dict,
-    Callable,
-    cast,
-    Union,
     Sequence,
-    Tuple,
     Set,
+    Tuple,
+    TypeVar,
+    Union,
+    cast,
 )
 from .wrappers import Limit, LimitGroup, RequestLimit
 
@@ -312,13 +312,15 @@ class Limiter:
                 self._storage_uri or storage_uri_from_config, **self._storage_options
             ),
         )
-        strategy = self._strategy or config.setdefault(
+        self._strategy = self._strategy or config.setdefault(
             ConfigVars.STRATEGY, "fixed-window"
         )
 
-        if strategy not in STRATEGIES:
-            raise ConfigurationError("Invalid rate limiting strategy %s" % strategy)
-        self._limiter = STRATEGIES[strategy](self._storage)
+        if self._strategy not in STRATEGIES:
+            raise ConfigurationError(
+                "Invalid rate limiting strategy %s" % self._strategy
+            )
+        self._limiter = STRATEGIES[self._strategy](self._storage)
 
         self._header_mapping = {
             HeaderNames.RESET: self._header_mapping.get(
@@ -351,7 +353,7 @@ class Limiter:
             ConfigVars.APPLICATION_LIMITS_COST, 1
         )
 
-        if not self.limit_manager.application_limits and app_limits:
+        if not self.limit_manager._application_limits and app_limits:
             self.limit_manager.set_application_limits(
                 [
                     LimitGroup(
@@ -377,7 +379,7 @@ class Limiter:
 
         conf_limits = config.get(ConfigVars.DEFAULT_LIMITS, None)
 
-        if not self.limit_manager.default_limits and conf_limits:
+        if not self.limit_manager._default_limits and conf_limits:
             self.limit_manager.set_default_limits(
                 [
                     LimitGroup(
@@ -403,7 +405,7 @@ class Limiter:
                 group.deduct_when = self._default_limits_deduct_when
                 group.cost = self._default_limits_cost
             self.limit_manager.set_default_limits(default_limit_groups)
-        self.__configure_fallbacks(app, strategy)
+        self.__configure_fallbacks(app, self._strategy)
 
         # purely for backward compatibility as stated in flask documentation
         if not hasattr(app, "extensions"):
@@ -544,29 +546,98 @@ class Limiter:
             cost=cost,
         )
 
+    @overload
     def exempt(
         self,
-        obj: Union[Callable[P, R], flask.Blueprint],
+        obj: flask.Blueprint,
+        *,
         flags: ExemptionScope = ExemptionScope.APPLICATION | ExemptionScope.DEFAULT,
-    ) -> Union[Callable[P, R], flask.Blueprint]:
+    ) -> flask.Blueprint:
+        ...
+
+    @overload
+    def exempt(
+        self,
+        obj: Callable[..., R],
+        *,
+        flags: ExemptionScope = ExemptionScope.APPLICATION | ExemptionScope.DEFAULT,
+    ) -> Callable[..., R]:
+        ...
+
+    @overload
+    def exempt(
+        self,
+        *,
+        flags: ExemptionScope = ExemptionScope.APPLICATION | ExemptionScope.DEFAULT,
+    ) -> Union[
+        Callable[[Callable[P, R]], Callable[P, R]],
+        Callable[[flask.Blueprint], flask.Blueprint],
+    ]:
+        ...
+
+    def exempt(
+        self,
+        obj: Optional[Union[Callable[..., R], flask.Blueprint]] = None,
+        *,
+        flags: ExemptionScope = ExemptionScope.APPLICATION | ExemptionScope.DEFAULT,
+    ) -> Union[
+        Callable[..., R],
+        flask.Blueprint,
+        Callable[[Callable[P, R]], Callable[P, R]],
+        Callable[[flask.Blueprint], flask.Blueprint],
+    ]:
         """
-        decorator to mark a view or all views in a blueprint as exempt from
+        Mark a view function or all views in a blueprint as exempt from
         rate limits.
 
-        :param obj: view or blueprint to mark as exempt.
+        :param obj: view function or blueprint to mark as exempt.
         :param flags: Controls the scope of the exemption. By default
          application wide limits and defaults configured on the extension
          are opted out of. Additional flags can be used to control the behavior
          when :paramref:`obj` is a Blueprint that is nested under another Blueprint
          or has other Blueprints nested under it (See :ref:`recipes:nested blueprints`)
+
+        The method can be used either as a decorator without any arguments (the default
+        flags will apply and the route will be exempt from default and application limits::
+
+            @app.route("...")
+            @limiter.exempt
+            def route(...):
+               ...
+
+        Specific exemption flags can be provided at decoration time::
+
+            @app.route("...")
+            @limiter.exempt(flags=ExemptionScope.APPLICATION)
+            def route(...):
+                ...
+
+        If an entire blueprint (i.e. all routes under it) are to be exempted the method
+        can be called with the blueprint as the first parameter and any additional flags::
+
+            bp = Blueprint(...)
+            limiter.exempt(bp)
+            limiter.exempt(
+                bp,
+                flags=ExemptionScope.DEFAULT|ExemptionScope.APPLICATION|ExemptionScope.ANCESTORS
+            )
+
         """
 
         if isinstance(obj, flask.Blueprint):
             self.limit_manager.add_blueprint_exemption(obj.name, flags)
-        else:
+        elif obj:
             self.limit_manager.add_route_exemption(
                 f"{obj.__module__}.{obj.__name__}", flags
             )
+        else:
+            _R = TypeVar("_R")
+            _WO = TypeVar("_WO", Callable[..., _R], flask.Blueprint)
+
+            def wrapper(obj: _WO) -> _WO:
+                return self.exempt(obj, flags=flags)
+
+            return wrapper
         return obj
 
     def request_filter(self, fn: Callable[[], bool]) -> Callable[[], bool]:
@@ -626,7 +697,10 @@ class Limiter:
 
     def check(self) -> None:
         """
-        check the limits for the current request
+        Explicitly check the limits for the current request. This is only relevant
+        if the extension was initialized with :paramref:`~flask_limiter.Limiter.auto_check`
+        set to ``False``
+
 
         :raises: RateLimitExceeded
         """
@@ -637,14 +711,25 @@ class Limiter:
         resets the storage if it supports being reset
         """
         try:
-            if self._storage:
-                self._storage.reset()
-                self.logger.info("Storage has been reset and all limits cleared")
+            self.storage.reset()
+            self.logger.info("Storage has been reset and all limits cleared")
         except NotImplementedError:
             self.logger.warning("This storage type does not support being reset")
 
+    @property
+    def storage(self) -> Storage:
+        """
+        The backend storage configured for the rate limiter
+        """
+        assert self._storage
+        return self._storage
+
     @property
     def limiter(self) -> RateLimiter:
+        """
+        Instance of the rate limiting strategy used for performing
+        rate limiting.
+        """
         if self._storage_dead and self._in_memory_fallback_enabled:
             limiter = self._fallback_limiter
         else:
@@ -784,28 +869,9 @@ class Limiter:
         if self.__check_all_limits_exempt(endpoint):
             return []
 
-        route_limits: List[Limit] = []
+        marked_for_limiting = name in self._marked_for_limiting
+        fallback_limits = []
 
-        before_request_context = in_middleware and name in self._marked_for_limiting
-
-        if not in_middleware and endpoint:
-            route_limits.extend(
-                self.limit_manager.route_limits(flask.current_app, endpoint)
-            )
-
-        if blueprint:
-            if not before_request_context and (
-                not route_limits
-                or all(not limit.override_defaults for limit in route_limits)
-            ):
-                route_limits.extend(
-                    self.limit_manager.blueprint_limits(flask.current_app, blueprint)
-                )
-
-        exemption_scope = self.limit_manager.exemption_scope(
-            flask.current_app, endpoint, blueprint
-        )
-        all_limits = []
         if self._storage_dead and self._fallback_limiter:
             if in_middleware and name in self._marked_for_limiting:
                 pass
@@ -819,25 +885,16 @@ class Limiter:
                     self._storage_dead = False
                     self.__check_backend_count = 0
                 else:
-                    all_limits = list(itertools.chain(*self._in_memory_fallback))
+                    fallback_limits = list(itertools.chain(*self._in_memory_fallback))
 
-        if not all_limits:
-            all_limits = (
-                self.limit_manager.application_limits
-                if in_middleware and not (exemption_scope & ExemptionScope.APPLICATION)
-                else []
-            )
-            all_limits += route_limits
-            explicit_limits_exempt = all(limit.method_exempt for limit in route_limits)
-            combined_defaults = all(
-                not limit.override_defaults for limit in route_limits
-            )
-
-            if (explicit_limits_exempt or combined_defaults) and not (
-                before_request_context or exemption_scope & ExemptionScope.DEFAULT
-            ):
-                all_limits += self.limit_manager.default_limits
-        return all_limits
+        return self.limit_manager.resolve_limits(
+            flask.current_app,
+            endpoint,
+            blueprint,
+            in_middleware,
+            marked_for_limiting,
+            fallback_limits,
+        )
 
     def __evaluate_limits(self, endpoint: str, limits: List[Limit]) -> None:
         failed_limits: List[Tuple[Limit, List[str]]] = []
